#include "D.h"
#include "sys_ctype.h"
#include "sys_dir.h"
#include "sys_stat.h"
#include "sys_errno.h"
#include "sys_fcntl.h"
#include "sys_time.h"
#include "sys_types.h"
#include "sys_sysmacros.h"
#include "sys_utime.h"
/*
* Copyright (C) 1993 by John Chambers <jc@trillian.mit.edu>.  All  rights  reserved.
* You  may  use  this program for any purpose whatsoever, as long as you don't
* claim that you wrote it (or that I wrote the changes that you made), and you
* agree to pass this source (with or without your changes) to anyone who asks.
*/
/*
* This is a generic copy/link/move/remove utility, with the ability to
* move files to "backup" rather than unlinking them.  It may be called
* via any of the names:
*
*   Cp [options] file1 file2	[Copy files]
*   cp [options] file1 file2
*   Cp [options] file... dir
*   cp [options] file... dir
*   Lc [options] file1 file2	[Link or copy files]
*   lc [options] file1 file2
*   Lc [options] file... dir
*   lc [options] file... dir
*   Ln [options] file1 file2	[Link files]
*   ln [options] file1 file2
*   Ln [options] file... dir
*   ln [options] file... dir
*   Mv [options] file1 file2	[Move files]
*   mv [options] file1 file2
*   Mv [options] file... dir
*   mv [options] file... dir
*
*   Rm [options] file...	[Remove files]
*   rm [options] file...
*   Bl [options] file...	[Break link]
*   bl [options] file...
*   Bp [options] file...	[Make backup copy]
*   bp [options] file...
*   Rs [options] file...	[Restore files]
*   rs [options] file...
*
*   clmr =<op> [options] args...
*
* This is a generic file utility that does moves, copies,  links,  and
* removal  of files.  When you use the upper-case names, files are not
* deleted; they are renamed. Several styles of renaming are supported,
* giving you several ways to remove old files to backup copies.
*
* This program is intended as  a  replacement  for  all  of  the  Unix
* library  programs with the same names.  Its command-line options and
* argument lists are the same as the Unix commands.  There are  a  few
* added  features.   One  "feature" is that this program's behavior is
* taken from its name in arg 0; if called via another name (mostly for
* testing purposes), you may use the '=' option as:
*    clmr =Cp args...
* to tell it how to behave.
*
* For the upper-case commands,  the  behavior  is  controlled  by  the
* BACKUP environment variable which is of the general form:
*   <n><c><s>
* where
*   <n> is a number limiting the backups to make,
*   <c> is a single character "backup flag",
*   <s> is a string to add as a suffix.
* If there is an initial number, it is the  limit  to  the  number  of
* backups  that will be made; the default is 8.  The <c> may be one of
* '-', '.', '*' or '#'.  The value of <s> is typically "bak" or "old";
* if used, you may be limited to only one backup. All three fields are
* optional; the default is "8-".  Examples:
*
* setenv BACKUP '7-'
*   This says to back up a file by appending '-' to its name.  This is
*   done recursively, so if "xx", "xx-" and "xx--" exist, they will be
*   renamed to "xx-", "xx--" and "xx---". The value of <n> should be a
*   number, which is the maximum number of such backups that  will  be
*   made. The default value for BACKUP is "8-", which will preserve up
*   to eight backup copies.
*
* setenv BACKUP '5#'
*   This  makes up to 5 backups by prepending '#' to the last field in
*   the file name. Thus "a/b/c" will be backed up to "a/b/#c", then to
*   "a/b/##c".  This takes somewhat more cpu time than the '-' suffix.
*
* setenv BACKUP '.bak'
* setenv BACKUP '-old'
*   These make backups by appending the suffix to the file name.  Thus
*   "a/b/c" will be renamed to "a/b/c.bak". The suffix must start with
*   '.' or '-' to be recognized, and any string may follow.  Only  one
*   backup copy may be made with this style.
*
* setenv BACKUP '5;'
*   This is the VMS backup scheme (which may not quite work yet).
*
* As with some (but not all) Unix versions, this program  assumes  "."
* as  a final arg if the final arg isn't a directory and the number of
* args is other than two.
*
* The options are a merger of those for cp, ln and mv, with POSIX's --
* thrown in for good measure.  You may use '+' instead of '-'  if  you
* prefer; there is no difference in behavior.
*
* A  "-"  alone  or "--" means that there are no more options, and all
* subsequent args are file names.  This is  one  of  several  ways  of
* making it possible for a file name to start with '-'.
*
* The -f option means "forced", and  tells  us  to  just  do  the  job
* silently,  without asking questions or harassing the user with silly
* error messages.  The  -F  option  is  from  Sys/V,  and  allows  the
* super-user  to force a hard link to a directory.  (This will produce
* an "illegal" file system in which the directories will no longer  be
* in  a  tree;  it is advisable to unlink the old directory as soon as
* possible).
*
* The -i option is used by cp and mv to get  "interactive"  operation.
* When  an  existing  file  would be overwritten, the user is told the
* file name, and we wait for a 'y' to tell us  to  do  it;  any  other
* answer  means "no".  (We also accept 'j' or 'o' as "yes", so you can
* answer in German or French if you like.  But note that if you answer
* in Greek, the sense will be reversed.  ;-)
*
* The -p option preserves times. Otherwise, for copies, new files will
* have the current timestamp.
*
* The -r option means "recursive".  If any of the source  files  is  a
* directory, it will be recursively copied (or moved or linked) to the
* target directory.  All needed directories will be created.
*
* The -s option is only meaningful when doing links; it means to do  a
* symlink (if possible).  Some Unix systems still don't have symlinks,
* but by now most do.
*
* This program has a few tricks  in  its  repertoire  in  addition  to
* mimicking the original Unix library programs.  Some are:
*
* The most useful new command is "lc" (or "Lc" to get backups).   This
* command  attempts  to  make  a link, and if that fails, does a copy.
* Thus lc may be used to install files in a remote directory when  you
* don't  know whether it's in the same file system.  Lc is very useful
* in the "install" entries of makefiles.
*
* Bp and bp are names that cause it to make a "backup" copy.  That is,
* it  results  in  both  the  original file and a backup version being
* present. Thus with the default BACKUP='-', if files "foo" and "foo-"
* exist, Bp will rename these to "foo-" "foo--", and also leave behind
* "foo" as it was originally.  Note that if "foo" was multiply  linked
* before,  then it will still be multiply linked; if it was a symlink;
* it will still be a symlink.  The difference between "bp" and "Bp" is
* that  "bp" makes (or overwrites) exactly one backup file; while "Bp"
* makes as many as BACKUP says to make, by renaming.
*
* Bl and bl are "break link" operations.  They result in a new, singly
* linked file.  Bl renames the original; bl unlinks it.  Thus if "foo"
* and "foo-" exist, "Bl foo" will rename them to "foo-"  and  "foo--",
* and then will do "cp -p foo- foo".  The "bl foo" command will create
* a new, singly-linked "foo" and copy the old file's data to  it;  the
* old file will disappear (and "foo-" will be unchanged).
*
* Rs and rs restore files from backup.  If given a backup name ("foo-"
* or  "##bar" or "qux.bak") that is consistent with the current BACKUP
* style, these commands will strip off the extry chars and reconstruct
* the basic file name. They will then look for a backup for that file,
* and copy ("Cp" or "cp" as appropriate) the backup to the file.  Thus
* you  can  use  either the desired file name or a backup name for the
* file, whichever is easiest.
*/

#define UNKNOWN 0
#define COPY    1
#define LINK    2
#define LNCP    3
#define MOVE    4
#define REMV    5
#define BLNK    6
#define BKUP    7
#define REST    8
#define OPCODES 8	/* Must equal highest opcode */

int  opcode   = 0;	/* Which operation are we doing? */
char *opname[]= {		/* For descriptive messages */
	unknown,"COPY","LINK","LNCP","MOVE","REMV","BLNK","BKUP","REST"};
char *opnm[]  = {		/* For short messages */
	null,"cp","ln","lc","mv","rm","bl","bp","rs"};

Str  buffer = {0};	/* Block buffer */
int  bufsiz = 10000;	/* Buffer size */

extern Flag backupfl;	/* Should we back up old files? */
extern Flag preserve;	/* Preserve file timestamps for cp/mv */

global Flag dosstyle = 0;	/* DOS-style command line */
extern int  exitstat;		/* Exit code, defined in V package now */
global Flag Forced   = 0;	/* Sys/V: force hard link to directory */
global Flag forced   = 0;	/* Just do it and keep quiet */
global Flag interact = 0;	/* Talk to the user about it */
global Flag moreopts = 1;	/* Process options */
global Flag msgs = 1;		/* Show error messages */
global Flag recurse  = 0;	/* Recursive */
global Flag symlinks = 0;	/* Symbolic or real links? */
global Flag vmsstyle = 0;	/* VMS-style command line */

global Flag intodir  = 0;	/* Moving things into a directory */
global Str  dirnam = {0};	/* Target directory name */

char on[]  = "on";
char off[] = "off";

char m_bkup[]   = "Back up %s";
char m_bkupfl[] = "Backup flag %s from c='%c' name=%s";
char m_called[] = "Called.";
char m_cld_cp[] = "Called \"%s\"";
char m_cld_tp[] = "Called for \"%s\" tp=%08X=\"%s\"";
char m_field[]  = "%d-byte field \"%s\"";
char m_ident[]  ="%s: %s and %s identical (not copied)";
char m_name[]   = "%d-byte name \"%s\"";
char m_nocopy[] = "%s: copy \"%s\" to \"%s\" failed [Err %d=%s=%s]";
char m_nolink[] = "%s: %slink \"%s\" to \"%s\" failed [Err %d=%s=%s]";
char m_nolndir[]= "%s: Can't link directories \"%s\" to \"%s\"";
char m_nomkdir[]= "%s: Can't make directory \"%s\" [Err %d=%s=%s]";
char m_noread[] = "%s: Can't read %s [Err %d=%s=%s]";
char m_normdir[]= "%s: Can't remove directory \"%s\" [Err %d=%s=%s]";
char m_nostat[] = "%s: Can't get status of %s [Err %d=%s=%s]";
char m_notyet[] = "### Not implemented yet ***";
char m_nounlnk[]= "%s: unlink \"%s\" failed [Err %d=%s=%s]";
char m_nowrite[]= "%s: Can't write %s [Err %d=%s=%s]";
char m_ntract[] = "Interacting; see if target exists ...";
char m_path[]   = "%d-byte path \"%s\"";
char m_rdir[]   = "Recurse thru directory \"%s\"";
char m_savtim[] = "Preserve timestamps for \"%s\"";
char m_srce[]   = "source: \"%s\"";
char m_targ[]   = "target: \"%s\"";
char m_target[] = "%d-byte target \"%s\"";
char m_unlink[] = "%s: Remove \"%s\"";

/*
* This routine returns:
*   1 if p is the name of a directory.
*   0 if p exists and is not a directory.
*  -1 if p doesn't exist or we can't stat() it.
*/
FCT isadir(p)
	char *p;
{	int  r = -1;
	STat st;
:3	Fenter("isadir");
	BZero(&st,sizeof(st));
:6	V6 m_cld_cp,p D;
	if ((r = Stat(p,&st)) < 0) {
:1		P3 m_nostat,pname,Dspp(p),Errinfo D;
		Fail;
	}
:5	V5 "Mode %08X=0%o for %s",st.st_mode,st.st_mode,Dspp(p) D;
	if (S_ISDIR(st.st_mode)) {
:5		V5 "Directory: \"%s\"",p D;
		r = 1;
	} else {
:5		V5 "Not a directory: \"%s\"",p D;
		r = 0;
	}
fail:
:3	Fexit;
	return r;
}

/*
* This routine returns:
*   1 if p is the name of a file.
*   0 if p doesn't exist or we can't stat() it.
*/
FCT isafile(p)
	char *p;
{	int  r;
	STat st;
:3	Fenter("isafile");
:6	V6 m_cld_cp,p D;
	r = (Stat(p,&st) == 0);
:3	Fexit;
	return r;
}

/*
* This routine examines p to see if it starts with a name for  any  of
* our  operations.   At  most two chars are examined.  The appropriate
* globals are set to indicate what is found.
*/
FCT pn(p)
	char *p;
{	int   r=0;
	int   c;
:3	Fenter("pn");
:6	V6 m_called D;
	switch (c = *p) {
	case 'B':
	case 'b':
		switch (p[1]) {
		case 'L': case 'l': opcode = BLNK; break;	/* Break links */
		case 'P': case 'p': opcode = BKUP; break;	/* Backup files */
		default: Fail;
		}
		break;
	case 'C':
	case 'c':
		switch (p[1]) {
		case 'L': case 'l':		/* cl* is a direct call */
			opcode = 0; break;	/* Not known (yet) */
		case 'O': case 'o':		/* Eventually do VMS-like "copy" */
			vmsstyle = 1; opcode = COPY; break;
		case 'P': case 'p': opcode = COPY; break;	/* Copy files */
		default: Fail;
		}
		break;
	case 'L':
	case 'l':
		switch (p[1]) {
		case 'C': case 'c': opcode = LNCP; break;	/* Link or copy files */
		case 'N': case 'n': opcode = LINK; break;	/* Link files */
		default: Fail;
		}
		break;
	case 'M':
	case 'm':
		switch (p[1]) {
		case 'O': case 'o':	/* Eventually do VMS-like "move" */
			vmsstyle = 1; opcode = MOVE; break;
		case 'V': case 'v': opcode = MOVE; break;	/* Move files */
		default: Fail;
		}
		break;
	case 'R':
	case 'r':
		switch (p[1]) {
		case 'M': case 'm': opcode = REMV; break;	/* Remove files */
		case 'S': case 's': opcode = REST; break;	/* Restore files */
		default: Fail;
		}
		break;
	}
	backupfl = ISupper(c);
:3	V3 m_bkupfl,backupfl?on:off,Dsp(c),Dspp(p) D;
fail:
:3	V3 "opcode=%d=%s from \"%s\"",opcode,opname[opcode],p D;
:3	Fexit;
	return r;
}

/*
* Print the usual sort of usage message.
*/
void usage()
{	int i;
	switch (opcode) {
	case COPY:
	case LINK:
	case LNCP:
	case MOVE:
		fprintf(stderr,"Usage: %s [options] file1 file2\n",pname);
		fprintf(stderr,"   or: %s [options] file... dir\n",pname);
		break;
	case REMV:
		fprintf(stderr,"Usage: %s [options] file...\n",pname);
		break;
	default:
		fprintf(stderr,"%s: Known names are",pname);
		for (i=1; i<=OPCODES; i++)
			fprintf(stderr," %s",opnm[i]);
		fprintf(stderr," or in upper case.\n");
	}
}

/*
* Interact with the user to find out whether we should destroy
* an existing file.  The caller has already checked to see that
* the file exists; we don't do that here.  We just pass its
* name to /dev/tty, and read the response.  If it starts with
* 'Y' or 'y', we return 1; else we return 0.
*/
FCT ia(name)
{	int r=1;
	int c;
:3	Fenter("ia");
:6	V6 "Called for file \"%s\"",NulC(name) D;
	Fprintf(stdout,"Overwrite %s? ",NulC(name));
	fflush(stdout);
	c = getc(stdin);
	r = (c == 'y') || (c == 'Y')	/* English "yes" */
	 || (c == 'j') || (c == 'J')	/* German "ja" */
	 || (c == 'o') || (c == 'O');	/* French "oui" */
:3	V5 "User answered '%c'",c D;
	while (c != '\n' && c != EOF)
		c = getc(stdin);
fail:
done:
:3	Fexit;
	return r;
}

/*
* Make Target.  Build a target name from the directory  name  and  the
* tail end of the pathname passed to us.
*/
FCT mt(path,tp)
	char *path;
	Str  *tp;
{	int  r=0;
	int  c, l, n;
	char *p, *q;
:3	Fenter("mt");
:6	V6 m_cld_tp,path,tp,tp?tp->v:null D;
	for (p=q=path; c = *q; q++) if (c == '/') p = q + 1;
	n = q - p;
	l = Strlen(tp->v);	/* Paranoia */
	if (l != tp->l) {
		V1 "### tp->l=%d changed to l=%d.",tp->l,l D;
		tp->l = l;
	}
:5	V5 m_name,n,p D;
	if (!MinStrM(tp,l+n+3,"mt-target")) Fail;
	tp->v[l] = '/';
	Strncpy(tp->v+1+l,p,n+1);
	tp->l += 1 + n;
:5	V5 m_target,tp->l,tp->v D;
	Done;
fail:
	exitstat = errno;
done:
:3	Fexit;
	return r;
}

/*
* This routine is intended to make the file a singly-linked file.   If
* it  is singly linked, nothing happens.  If it is multiply-linked, it
* is renamed, a copy is made to the original  name,  and  the  renamed
* file is unlinked.  The timestamps should be copied, too.
*/
FCT bl(name)
	char *name;
{	int  r=0;
:3	Fenter("bl");
:6	V6 m_cld_cp,name D;
	V1 m_notyet D;
:3	Fexit;
	return r;
}

/*
* Here's the routine to do a file copy.  This always means reading the
* source  file  and  writing the target, possibly after backing it up.
* There's some cruftiness to avoid the problem of  trying  to  copy  a
* file to itself, resulting in an empty file: we get the status blocks
* for both files and if they are identical, we refuse to copy.   Also,
* note  the df arg, which tells us whether to consider the possibility
* that the target may be a directory.  The mv() and lc() routines pass
* df=0  to  say  that they've already done the mt() call to append the
* file name to a directory name, and cp() shouldn't do it again.
*/
FCT cp(name,tp,df)
	char *name;
	Str  *tp;
	Flag  df;	/* True if tp may be a directory */
{	int   r=0;
	int   f1, f2;
	char *n1, *n2;
	int   i, n;
	int   fcount, offset, rcount, wcount;
	size_t bytes=0;
	Flag  gots1=0, gots2=0;
	STat  s1, s2;
#ifdef USE_utime
	struct utimbuf tb;
#endif
#ifdef USE_utimes
	TIMV  tv[2];
#endif
:3	Fenter("cp");
:6	V6 m_cld_tp,name,tp,tp?tp->v:null D;
:3	V3 "Copy \"%s\" to \"%s\"",name,tp?tp->v:null D;
	if (Stat(n1=name,&s1) == 0) {	/* Does source exist? */
		gots1 = 1;
	} else {
		exitstat = errno;
:1		P2 m_nostat,pname,Dspp(n1),Errinfo D;
		Fail;
	}
:5	V5 "df=%d intodir=%d.",df,intodir D;
	if (df && intodir)
		mt(name,tp);	/* Calculate target name */
	if (Stat(n2=tp->v,&s2) == 0) {	/* Does target exist? */
		gots2 = 1;
	}
:3	V5 m_srce,n1 D;
:3	V5 m_targ,n2 D;
	if (backupfl) {
:3		V5 m_bkup,Dspp(n2) D;
		r = Backup(n2);
:1		if (r < 0) {
:1			if (msgs || Vlvl>1)
:1				P0 "%s: Can't backup \"%s\" [Err %d=%s=%s]",pname,n2,Errinfo D;
:1		}
	} else
	if (interact) {
:3		V5 m_ntract D;
		if (interact && isafile(n1))
			if (!ia(n1)) Fail;
	}
	if (gots2 && !backupfl) {	/* Check for source == target */
		if (BEq(&s1,&s2,sizeof(STat))) {
:2			P2 m_ident,pname,Dspp(n1),Dspp(n2) D;
			Fail;
		}
	}
	if ((f1 = Open(n1,0,0)) < 0) {
		if (msgs || Vlvl>1)
			P0 m_noread,pname,Dspp(n1),Errinfo D;
		r = -1;
		exitstat = errno;
		Fail;
	}
:3	if (preserve) {
:3		V5 m_savtim,n1 D;
:3	}
	if ((f2 = Open(n2,O_WRONLY|O_CREAT|O_TRUNC,s1.st_mode)) < 0) {
		exitstat = errno;
:1		if (msgs || Vlvl>1)
:1			P0 m_nowrite,pname,Dspp(n2),Errinfo D;
		Close(f1);
		r = -1;
		Fail;
	}
	if (!buffer.v) {
		if (!MinStrM(&buffer,bufsiz+2,"buffer")) {
			exitstat = ENOMEM;
			Fail;
	}	}
	if (preserve) {
:3		V5 m_savtim,n1 D;
		if (Fstat(f1,&s1) == 0) {
			gots1 = 1;
:2		} else {
:2			P1 m_nostat,pname,Dspp(n1),Errinfo D;
		}
	}
	while ((rcount = Read(f1,buffer.v,bufsiz)) > 0) {
		fcount = offset = 0;
loop:
		if ((wcount = Write(f2,buffer.v+offset,rcount)) < rcount) {
:3			V3 "Write gave %d < %d [Err %d=%s=%s]",wcount,rcount,Errinfo D;
			switch (errno) {
#if defined(EDQUOT)
			case EDQUOT:
#endif
#if defined(ENOSPC)
			case ENOSPC:
#endif
#if defined(EDQUOT) || defined(ENOSPC)
:1				if (fcount++ % 10)	/* Complain every 10 seconds */
:1					V1 "Blocked writing \"%s\" [Err %d=%s=%s]",n2,Errinfo D;
				if (wcount > 0) {
					offset += wcount;	/* Don't rewrite the stuff that worked */
					rcount -= wcount;
					bytes += wcount;
				}
				Sleep(1);
				Loop;
#endif
			default:
				exitstat = errno;
				if (msgs || Vlvl>1)
					P0 "%s: Write to \"%s\" failed [Err %d=%s=%s]",
						pname,n2,Errinfo D;
				Close(f2);
				Close(f1);
				r = -1;
				Fail;
			}
		}
		bytes += wcount;
	}
	if (rcount >= 0) {
		r = bytes;
:3		V5 "Copied %d bytes.",r D;
:2	} else {
:2		V1 "Read \"%s\" failed [Err %d=%s=%s]",n1,Errinfo D;
	}
	if (preserve && gots1) {
#ifdef USE_utime
		tb.actime  = s1.st_atime;
		tb.modtime = s1.st_mtime;
		utime(n2,&tb);
#define SET_times
#endif
#ifdef USE_utimes
		tv[0].tv_sec = s1.st_atime;
		tv[1].tv_sec = s1.st_mtime;
		tv[0].tv_usec = tv[1].tv_usec = 0;
		utimes(n2,tv);
#define SET_times
#endif
#ifndef SET_times
:7		V2 "Don't know how to set file times on this system." D;
#endif
	}
	Close(f2);
	Close(f1);
fail:
:3	Fexit;
	return r;
}

/*
* Link name to tp; if that fails because they are on different file
* systems, do a copy instead.
*/
FCT lc(name,tp,df)
	char *name;
	Str  *tp;
	Flag  df;	/* True if tp may be a directory */
{	int  r=0;
	char *n1, *n2;
	int  tl;
:3	Fenter("lc");
:6	V6 m_cld_tp,name,tp,tp?tp->v:null D;
:3	V4 "Link or copy \"%s\" to \"%s\"",name,tp?tp->v:null D;
:2	if (!tp) {
:2		V2 "### Called with null target tp=%08X tp->v=%08X",tp,tp?tp->v:null D;
:2		Fail;
:2	}
	tl = tp->l;
:5	V5 "df=%d intodir=%d.",df,intodir D;
	if (df && intodir)	/* Is target a directory? */
		mt(name,tp);		/* Calculate target name */
	n1 = name;
	n2 = tp->v;
:3	V5 m_srce,n1 D;
:3	V5 m_targ,n2 D;
	if (backupfl) {
:3		V5 m_bkup,Dspp(n2) D;
		Backup(n2);
	} else
	if (interact && isafile(n2))
		if (!ia(n2)) Fail;
	if (symlinks) {
:3		V3 "symlink \"%s\" to \"%s\"",n1,n2 D;
		r = symlink(n1,n2);
	} else
	if (isadir(n1) && isadir(n2) && !(Forced || symlinks)) {
:2		P2 m_nolndir,pname,n1,n2 D;
		Fail;
	} else {
:3		V3 "link \"%s\" to \"%s\"",n1,n2 D;
		r = Link(n1,n2);
	}
	if (r < 0) {
:1		P2 m_nolink,pname,symlinks?"sym":null,n1,n2,Errinfo D;
		if (isadir(n1)) {
			exitstat = EISDIR;
:2			P2 "%s: Can't link directory \"%s\" to \"%s\"",pname,n1,n2 D;
			Fail;
		} else {
			n2[tp->l = tl] = 0;	/* Strip off stuff added by mt above */
			if (cp(n1,tp,df) < 0) {		/* Try to cp it instead */
				exitstat = errno;
:1				P2 m_nocopy,pname,n1,n2,Errinfo D;
				r = -1;
				Fail;
			}
		}
	}
fail:
:3	Fexit;
	return r;
}

/*
* Link name to tp, which may be a directory.
*/
FCT ln(name,tp)
	char *name;
	Str  *tp;
{	int  r=0;
	char *n1, *n2;
	int  tl;
:3	Fenter("ln");
:6	V6 m_cld_tp,name,tp,tp?tp->v:null D;
:3	V5 "Link \"%s\" to \"%s\"",name,tp?tp->v:null D;
	if (intodir) mt(name,tp);
	n1 = name;
	n2 = tp->v;
:3	V5 m_target,tp->l,tp->v D;
	if (backupfl) {
:3		V5 m_bkup,Dspp(n2) D;
		Backup(n2);
	} else
	if (isadir(n1) && isadir(n2) && !(Forced || symlinks)) {
:1		P2 m_nolndir,pname,n1,n2 D;
		Fail;
	}
	if (interact && isafile(n2))
		if (!ia(n2)) Fail;
	if (symlinks) {
:3		V3 "symlink \"%s\" to \"%s\"",n1,n2 D;
		r = symlink(n1,n2);
	} else {
:3		V3 "link \"%s\" to \"%s\"",n1,n2 D;
		r = Link(n1,n2);
	}
	if (r < 0) {
:1		P2 m_nolink,pname,symlinks?"sym":null,n1,n2,Errinfo D;
		exitstat = errno;
	}
fail:
:3	Fexit;
	return r;
}

/*
* Here's the "mv" command.  Note that we may get a directory name, for
* the  non-recursive  case.  In the recursive case, we should only get
* the files, and dr() should take  care  of  deleting  the  source
* directories.
*/
FCT mv(name,tp)
	char *name;
	Str  *tp;
{	int  r=0;
	char *n1, *n2;
	Flag gots1=0, gots2=0;
	STat s1, s2;
:3	Fenter("mv");
:6	V6 m_cld_tp,name,tp,tp?tp->v:null D;
:3	V3 "Move \"%s\" to \"%s\"",name,tp?tp->v:null D;
	if (Stat(n1=name,&s1) == 0) {	/* Does source exist? */
		gots1 = 1;
	} else {
		exitstat = errno;
:1		P2 m_nostat,pname,Dspp(n1),Errinfo D;
		Fail;
	}
	if (intodir) mt(name,tp);	/* Calculate target name */
	if (Stat(n2=tp->v,&s2) == 0) {	/* Does target exist? */
		gots2 = 1;
	}
:3	V5 m_target,tp->l,tp->v D;
	if (backupfl) {
:3		V5 m_bkup,Dspp(tp->v) D;
		Backup(tp->v);
	} else
	if (interact && isafile(tp->v))
		if (!ia(tp->v)) Fail;
	if (gots2 && !backupfl) {	/* Check for source == target */
		if (BEq(&s1,&s2,sizeof(STat))) {
:2			P2 m_ident,pname,Dspp(n1),Dspp(n2) D;
			Fail;
		}
	}
/*
* First, we try the rename() system call.  If it works, it's the  best
* way to do it. But it doesn't always work right on some Unix systems.
*/
	if (rename(name,tp->v) == 0) {
:3		V3 "Rename \"%s\" to \"%s\"",name,tp->v D;
		Done;
:2	} else {
:2		V3 "Rename \"%s\" to \"%s\" failed [Err %d=%s=%s]",
:2			name,tp->v,Errinfo D;
	}
/*
* If that failed, we try making a link and then unlinking the old
* one.  This will fail for directories (unless we are superuser on
* the rare systems that allow superusers to link directories).
*/
	if (Link(name,tp->v) < 0) {
:1		P2 m_nolink,pname,null,name,tp->v,Errinfo D;
		switch (errno) {
		case EACCES:	/* If link gets these, read/write will too */
		case ENOENT:
			exitstat = errno;
			Fail;
		}
		if (cp(name,tp,0) < 0) {
			exitstat = errno;
:1			P2 m_nocopy,pname,name,tp->v,Errinfo D;
			r = -1;
			Fail;
		}
	}
	if (Unlink(name) < 0) {
		exitstat = errno;
		r = -1;
:1		if (msgs || Vlvl>1)
:1				P0 m_nounlnk,pname,name,Errinfo D;
	}
done:
fail:
:3	Fexit;
	return r;
}

/*
* This routine merely removes the named file.
*/
FCT rm(name)
	char *name;
{	int  r=0;
:3	Fenter("rm");
:6	V6 m_cld_cp,name D;
:3	P3 m_unlink,pname,name D;
	if (backupfl) {
:3		V5 m_bkup,Dspp(name) D;
		Backup(name);
		Done;
	}
	if (interact && isafile(name))
		if (!ia(name)) Fail;
	if (Unlink(name) < 0) {
		r = -1;
:1		if (msgs || Vlvl>1)
:1			P0 m_nounlnk,pname,name,Errinfo D;
	}
fail:
done:
:3	Fexit;
	return r;
}

static Str target = {0};

main(ac,av)
	char **av;
{	int   a, i, l, n;
	int   c0, c1;
	char *p;
	char *lastarg;
	STat  st;
	Str  *tp=0;

:9	printf(Started.");
:9	fflush(stdout);
:9	printf(Started \"%s\"",av[0]);
:9	fflush(stdout);
:1	ac = Dinit(ac,av);
	if (p = Getenv("BACKUP")) bpoption(p);
loop:
:3	V4 "Get operation from pname=\"%s\"",pname D;
	pn(pname);
	for (a=n=1; a<ac; a++)
:5		V5 "Arg %2d: \"%s\"",a,av[a] D;
:5	V5 "Run thru args 1-%d to get options ...",ac-1 D;

	for (a=n=1; a<ac; a++) {
:5		V5 "Arg %2d: \"%s\"",a,av[a] D;
		if (moreopts)
		switch (c0 = av[a][0]) {
		case '=':	/* Explicit function */
			av += a;	/* Make this the first arg */
			ac -= a;	/* Adjust the arg count */
			av[0]++;	/* Skip over the '=' */
			pname = av[0];
:2			V2 "Command is %s ...",Dspp(pname) D;
			Loop;		/* Start over with this arg */
		case '-':
		case '+':
:5			V5 "Opt %2d: \"%s\"",a,av[a] D;
			for (l=1; c1=av[a][l]; l++)
			switch (c1) {
			/*
			* a "-" alone or "--" means that there are no more
			* options, and all subsequent args are file names.
			* This is one of several ways of making it possible
			* for a file name to start with '-'.
			*/
			case '-': case  0 :
				moreopts = 0;
:3				V4 "Nor more options." D;
				break;
			/*
			* The +b option means to make backups.  If followed by
			* anything, the rest of the option is taken as the backup
			* style.  This will override the BACKUP environment variable.
			*/
			case 'b':
			case 'B':
				if (backupfl = (c0 == '+')) {
:5					V5 "Making backups." D;
					if (av[a][2]) {
:5						V5 "There's a backup option \"%s\"",av[a]+2 D;
						bpoption(av[a]+2);
					}
				}
:3				V3 "Backups %sabled.",backupfl?"en":"dis" D;
				break;
			/*
			* The -f option means "forced", and tells us to just do
			* the job silently, without asking questions or harassing
			* the user with silly error messages.  The -F option is
			* from Sys/V, and allows the super-user to force a hard
			* link to a directory.  (This will produce an "illegal"
			* loop and the directories will no longer be in a tree;
			* it is advisable to unlink the old directory as soon as
			* possible).
			*/
			case 'F':
				Forced = 1;	/* Sys/V: force hard link to directory */
:3				V3 "Forced directory link." D;
			case 'f':
				forced = 1;
				msgs = 0;
:3				V3 "Forced." D;
				break;
			/*
			* The -i option is used by cp and mv to get "interactive"
			* operation.  When an existing file would be overwritten,
			* the user is told the file name, and we wait for a "yes"
			* to tell us to do it.
			*/
			case 'I': case 'i':
				interact = 1;
:3				V3 "Interact." D;
				break;
			/*
			* The -p option preserves times.  Otherwise, for copies,
			* new files will have the current timestamp.
			*/
			case 'P': case 'p':
				preserve = 1;
:3				V3 "Preserve times." D;
				break;
			/*
			* The -r option means "recursive".  If any of the source
			* files is a directory, it is recursively copied (or moved
			* or linked) to the target.  All needed directories will
			* be created.
			*/
			case 'R': case 'r':
				recurse = 1;
:3				V3 "Recursive." D;
				break;
			/*
			* The -s option is only meaningful when doing links; it
			* means to do a symlink (if possible).  Some Unix systems
			* still don't have symlinks, but by now most do.
			*/
			case 'S': case 's':
				symlinks = 1;
				break;
			default:
:1				V1 "Unknown option \"%s\" ignored.",av[a] D;
				break;
			}
			continue;
		}
:3		V4 "File arg \"%s\" ignored in option pass.",av[a] D;
		av[n++] = av[a];
	}
	moreopts = 0;
:3	V5 "There are %d args left.",n-1 D;
	av[ac=n] = 0;	/* Paranoia */
	if (!opcode) {
:1		V1 "can't determine operation from pname=\"%s\"",pname D;
		exitstat = -1;
		Fail;
	}
:3	V5 "Operation: %d=%s=%s.",opcode,opnm[opcode],opname[opcode] D;
	switch (opcode) {
	case COPY:
	case LINK:
	case LNCP:
	case MOVE:
		lastarg = av[ac-1];
:3		V4 "Is the final arg %d \"%s\" a directory?",ac-1,lastarg D;
		if (isadir(lastarg) > 0) {
			dirnam.v = av[--ac];
			dirnam.l = strlen(dirnam.v);
			intodir = 1;
:3			V3 "Target directory: \"%s\"",dirnam.v D;
		} else
		if (ac == 3) {
			V4 "Only two args (intodir=%d)",intodir D;
			target.v = av[--ac];
			target.l = strlen(target.v);
		} else
		if (isafile(lastarg) > 0) {
:2			P2 "%s: Wrong number of args %d; adding \".\".",pname,ac-1 D;
			dirnam.v = ".";	/* Default to current dir */
			dirnam.l = 1;
			intodir = 1;
		} else {
:2			P2 "%s: Wrong number of args %d; \"%s\" doesn't exist.",
:2				pname,ac-1,lastarg D;
			usage();
			exitstat = -1;
			Fail;
		}
		break;
	case BKUP:
	case BLNK:
	case REMV:
	case REST:
		break;
	}
:3	V4 "The intodir flag is %s.",intodir?on:off D;
:6	V6 "The dos-style flag is %s.",dosstyle?on:off D;
:6	V6 "The vms-style flag is %s.",vmsstyle?on:off D;
:5	V5 "Run thru args 1-%d to get files ...",ac-1 D;

	for (a=1; a<ac; a++) {
:5		V5 "arg %2d: \"%s\"",a,av[a] D;
		if (recurse && (isadir(av[a]) > 0)) {
:3			V4 m_rdir,av[a] D;
			dr(av[a],&dirnam);
			continue;
		}
		if (!tp && dirnam.v) {
			tp = &dirnam;
:5			V5 "tp=%08X [dirnam] \"%s\"",tp,tp->v D;
		}
		if (!tp && target.v) {
			tp = &target;
:5			V5 "tp=%08X [target] \"%s\"",tp,tp->v D;
		}
		fl(av[a],tp);
:5		V5 "arg %2d: \"%s\" done.",a,av[a] D;
	}
fail:
:3	V3 "Exit with status %d.",exitstat D;
	Exit(exitstat);
}

/*
* This routine has the job of running thru the  named  directory,  and
* generating  a list of source file paths and the corresponding target
* names for each.  It then invokes fl() to do the correct job  on
* each source/target pair.
*/
FCT dr(name,tp)
	char *name;
	Str  *tp;
{	int  r=0;
	int  c, i, l, n;
	Str  path;
	Str  target;	/* Target name */
	char *p, *q;
	DIR *dirp;
	DIRE*dp;
:3	Fenter("dr");
	ZeroStr(path);
	ZeroStr(target);
:5	V6 m_cld_tp,name,tp,tp?tp->v:null D;

	for (p=q=name; c = *q; q++) if (c == '/') p = q + 1;
	l = q - p;	/* Length of final field of name */
:3	V5 m_field,l,p D;
/*
* Now we have p -> the final field of name, and l its length.  We will
* need to generate a target name consisting of tp->v/p/foo for each
* foo in the named directory.
*/
	if (!(dirp = opendir(name))) {
:1		if (msgs || Vlvl>1)
:1			P0 m_noread,pname,Dspp(name),Errinfo D;
		Fail;
	}
	n = strlen(name);
	if (!MinStrM(&path,n+17,"dirent"))
			Fail;
	Strncpy(path.v,name,n);
	path.v[n++] = '/';
	path.v[path.l = n] = 0;
:3	V5 m_path,path.l,path.v D;

	switch (opcode) {
	case REMV:
		break;
	default:
:5		V5 "Make %d-byte target name from tp=\"%s\" and p=\"%s\"",n+l+3,tp->v,p D;
		if (!MinStrM(&target,n+l+3,"dr-target")) Fail;
		Strncpy(target.v,tp->v,tp->l);
		target.v[tp->l] = '/';
		Strncpy(target.v+tp->l+1,p,l+1);
		target.l = tp->l + 1 + l;
:3		V5 m_target,target.l,target.v D;
		if (MkDir(target.v,0777) < 0) {
			P2 m_nomkdir,pname,target.v,Errinfo D;
			/* We barge ahead anyway ... */
		}
		break;
	}
/*
* We have set up path to contain the input directory name and a '/',
* with n the offset to just after the '/'.  We now need to run thru
* the named directory and extract each file name.
*/
	for (dp = readdir(dirp); dp; dp = readdir(dirp)) {
:3		V5 "Dirent: \"%s\"",dp->d_name D;
		if (dp->d_name[0] == '.') {
:3			V6 "Hidden: \"%s\"",dp->d_name D;
			if (!dp->d_name[1]) {
:3				V5 "Ignore: \"%s\"",dp->d_name D;
				continue;
			}
			if (dp->d_name[1] == '.' && !dp->d_name[2]) {
:3				V5 "Ignore: \"%s\"",dp->d_name D;
				continue;	/* Ignore ".." */
			}
		}
		i = strlen(dp->d_name) + 1;	/* Length of dirent name + null */
		if (n + i >= path.m)
			if (!MinStrM(&path,n+i+1,"dirent"))
				Fail;
		Strncpy(path.v+n,dp->d_name,i);
:5		V5 m_path,  path.l,  path.v D;
:5		V5 m_target,target.l,NulC(target.v) D;
		if (recurse && (isadir(path.v) > 0)) {
:3			V5 m_rdir,path.v D;
			dr(path.v,&target);
			continue;
		}
		fl(path.v,&target);
		if (target.v)
			target.v[target.l = tp->l + 1 + l] = 0;
	}
	closedir(dirp);
	switch (opcode) {	/* Cases where we remove the source directory */
	case MOVE:
	case REMV:
		if (RmDir(name) < 0) {
			if (msgs || Vlvl>1)
				P0 m_normdir,pname,name,Errinfo D;
			Fail;
		}
	}
fail:
	if (path.v) ZapStrM(&path,"path");	/* Free any malloced space */
	if (target.v) ZapStrM(&target,"target");
:3	Fexit;
	return r;
}

/*
* Here's a routine to handle the processing of  one  file.   The  main
* thing  we do here is to decide which of the action routines to call,
* and then to restore the target name afterwards.  We need to be a bit
* careful, because for rm, there won't * be a target.
*/
FCT fl(name,tp)
	char *name;
	Str  *tp;
{	int  r=0;
	int  tl=0;
:3	Fenter("fl");
:5	V6 m_cld_tp,name,tp,tp?tp->v:null D;
	if (tp) tl = tp->l;

	switch (opcode) {
	case COPY: cp(name,tp,1); break;
	case LINK: ln(name,tp); break;
	case LNCP: lc(name,tp,1); break;
	case MOVE: mv(name,tp); break;
	case REMV: rm(name); break;
	case BKUP: Backup(name); break;
	case REST: rs(name); break;
	case BLNK: bl(name); break;
:2	default:
:2		V2 "### Unknown opcode %d.",opcode D;
	}
	if (tp && tp->v) {	/* Discard anything added to target */
:5		V5 "Cut \"%s\" from %d to %d bytes.",tp->v,tp->l,tl D;
		tp->v[tp->l = tl] = 0;
:5		V5 "Got \"%s\"",tp->v D;
	}
:3	Fexit;
	return r;
}
